"""


"""
from tropicpy.attacks.attacks_utils import *


def kotov_ushakov_simple(A, B, g, U, V):
    for i in range(1, g + 1):
        for j in range(1, g + 1):
            print("T_" + str(i) + str(j) + "=\n" + str(matrix_difference(U, (A ** i) * (B ** j))))
            c = is_matrix_difference_constant(U, (A ** i) * (B ** j))
            if c:
                print("Attack was succesfull!")
                print("c = " + str(c))
                print("i = " + str(i))
                print("j = " + str(j))

                return ((A ** i) * c) * V * (B ** j)
    return False



"""


"""
from tropicpy.tropical.tropical_utils import *

inf = "inf"


class TropicalValue(object):
    def __init__(self, value):
        if value == inf or value == float(inf):
            self.is_num = False
            self.value = float(inf)
        elif is_number(value):
            self.is_num = True
            self.value = value
        else:
            raise Exception(str(value) + " is not an accurate tropical value.")

    def __str__(self):
        return str(self.value)

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        return TropicalValue(min(self.value, other.value))

    def __mul__(self, other):
        return TropicalValue(self.value + other.value)

    def __pow__(self, power, modulo=None):
        return self.value * power


tropical_0 = TropicalValue(inf)
tropical_1 = TropicalValue(0)


def correct_tropical_value(value):
    try:
        TropicalValue(value)
    except:
        return False
    else:
        return True



"""


"""
from tropicpy.attacks.attacks_utils import *


def kotov_ushakov(A, B, g, U, V, p_min):
    n = A.rows
    data = []

    for i in range(g + 1):
        for j in range(g + 1):
            matrix_min = min_of_matrix_difference((A ** i) * (B ** j), U * TropicalValue((-2) * -1000))
            if matrix_min[0].value <= 0:
                result = {"i": i, "j": j, "minval": matrix_min[0], "inds": matrix_min[1]}
                data.append(result)

    inds = [result["inds"] for result in data]
    C = minimal_set_covers(inds)

    minimal_covers_ij = []
    for cover in C:
        minimal_covers_ij.extend(
            list_cartesian([[[record["i"], record["j"]] for record in data if el == record["inds"]] for el in cover]))

    for ij in minimal_covers_ij:
        result = simplex(make_simplex_matrix(data, g, ij))
        if result:
            polys = [[int(result[i] + p_min) for i in range(g + 1)], [int(result[g + 1 + i] + p_min) for i in range(g + 1)]]
            p1_A = tropical_matrix_0(n)
            p2_B = tropical_matrix_0(n)
            for (el1, el2, i) in zip(polys[0], polys[1], range(g + 1)):
                p1_A += (A ** i) * TropicalValue(el1)
                p2_B += (B ** i) * TropicalValue(el2)

            print("Attack was succesfull!")
            print("p1(A)=\n" + str(p1_A))
            print("p2(B)=\n" + str(p2_B))

            return p1_A * V * p2_B

    return False



"""


"""

from tropicpy.tropical.tropical_value import *
import random

class TropicalMatrix:
    def __init__(self, values):

        tmp_rows = len(values)
        if tmp_rows == 0:
            tmp_columns = 0
        else:
            tmp_columns = len(values[0])
        tmp_values = []

        for (row, i) in zip(values, range(tmp_rows)):
            if len(row) != tmp_columns:
                raise Exception("Missing values in row no. " + str(i + 1) + ": " + str(row))

        for (row, i) in zip(values, range(tmp_rows)):
            tmp_row = []
            for value in row:
                if correct_tropical_value(value):
                    tmp_row.append(TropicalValue(value))
                elif isinstance(value, TropicalValue) and correct_tropical_value(value.value):
                    tmp_row.append(value)
                else:
                    raise Exception(str(value) + " is not an accurate element of a tropical matrix.")
            tmp_values.append(tmp_row)

        self.rows = tmp_rows
        self.columns = tmp_columns
        self.values = tmp_values

    def __str__(self):
        if len(self.values) == 0:
            return "[]"

        result = ""
        column_widths = []
        for col_no in range(self.columns):
            column_widths += [max(len(str(row[col_no])) for row in self.values)]

        for (row, row_no) in zip(self.values, range(self.rows)):
            result += "["
            for (value, col_no) in zip(row, range(self.columns)):
                result += '{0:^{1}}'.format(str(value), column_widths[col_no] + 2)
            result += "]"
            if row_no != self.rows - 1:
                result += "\n"
        return result

    def __eq__(self, other):
        if isinstance(other, TropicalMatrix):
            if self.rows != other.rows:
                return False
            elif self.columns != other.columns:
                return False

            for (row_a, row_b) in zip(self.values, other.values):
                for (a, b) in zip(row_a, row_b):
                    if a != b:
                        return False
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        if self + other == self:
            return True
        else:
            return False

    def __add__(self, other):
        if self.rows != other.rows or self.columns != other.columns:
            raise Exception("Different dimensions of matrices.")

        result = []
        for (row_a, row_b) in zip(self.values, other.values):
            tmp_row = []
            for (a, b) in zip(row_a, row_b):
                tmp_row.append(a + b)
            result.append(tmp_row)
        return TropicalMatrix(result)

    def __mul__(self, other):
        if isinstance(other, TropicalMatrix):
            if self.columns != other.rows:
                raise Exception("Mismatched dimensions of matrices.")

            result = []
            other_columns = []

            for col_no in range(other.columns):
                other_columns.append([row[col_no] for row in other.values])

            for row in self.values:
                tmp_row = []
                for col_no in range(self.columns):
                    times = list(map(TropicalValue.__mul__, row, other_columns[col_no]))
                    tropical_sum = tropical_0
                    for el in times:
                        tropical_sum += el
                    tmp_row.append(tropical_sum)
                result.append(tmp_row)
            return TropicalMatrix(result)
        elif isinstance(other, TropicalValue):
            result = []
            for row in self.values:
                tmp_row = []
                for value in row:
                    tmp_row.append(other * value)
                result.append(tmp_row)

            return TropicalMatrix(result)
        else:
            raise Exception("Cannot perform tropical multiplication for type: " + str(type(other)))

    def __pow__(self, power, modulo=None):
        if isinstance(power, int):
            if power == 0:
                return tropical_matrix_1(self.rows)
            elif power == 1:
                return self
            elif power % 2 == 0:
                return (self * self) ** (power >> 1)
            else:
                return self * ((self * self) ** ((power - 1) >> 1))
        else:
            raise Exception(str(power) + " is not an accurate power.")

    def __matmul__(self, other):
        if isinstance(other, TropicalMatrix):
            return self + other + (self * other)
        else:
            raise Exception("Cannot perform adjoint multiplication for type: " + str(type(other)))

    def __xor__(self, power):
        if power == 1:
            return self
        elif power % 2 == 0:
            return (self @ self) ^ (power >> 1)
        else:
            return self @ ((self @ self) ^ ((power - 1) >> 1))


def semidirect_product_1st(A, B, C, D):
    return (A @ D) + C


def semidirect_product_2nd(A, B, C, D):
    return B @ D


# originally square-and-multiply method
def semidirect_power_1st(A, B, n):
    if n == 1:
        return A
    else:
        I = tropical_matrix_1(A.rows)
        return ((A * (I + B)) + B) * ((I + B) ** (n - 2))


def semidirect_power_2nd(A, B, n):
    return B ^ n


def tropical_matrix_0(n):
    values = []
    for i in range(n):
        values.append([tropical_0] * n)
    return TropicalMatrix(values)


def tropical_matrix_1(n):
    values = []
    for i in range(n):
        tmp_row = [tropical_0] * n
        tmp_row[i] = tropical_1
        values.append(tmp_row)
    return TropicalMatrix(values)


def generate_random_tropical_value(l, u):
    value = random.randint(l, u)
    if random.randint(1, 100) == 1:
        value = inf
    return TropicalValue(value)


def generate_random_tropical_matrix(n, l, u):
    values = []
    for i in range(n):
        tmp_row = []
        for j in range(n):
            tmp_row.append(generate_random_tropical_value(l, u))
        values.append(tmp_row)
    return TropicalMatrix(values)



"""


"""
from tropicpy.tropical.tropical_matrix import *


class GrigorieShpilrain2014:

    def __init__(self, A, B, g, p_min, p_max):
        if not isinstance(A, TropicalMatrix):
            raise Exception(str(A) + " is not an appropriate value.")
        elif not isinstance(B, TropicalMatrix):
            raise Exception(str(B) + " is not an appropriate value.")
        else:
            if A.rows != B.rows or A.columns != B.columns:
                raise Exception("Matrices A and B are of different dimensions.")
            elif A.rows != A.columns:
                raise Exception("Matrix A is not a square matrix.")
            elif B.rows != B.columns:
                raise Exception("Matrix B is not a square matrix.")
            elif A * B == B * A:
                raise Exception("Matrices A and B do not satisfy the condition A*B!=B*A.")

            self.n = A.rows
            self.A = A
            self.B = B
            self.g = g

            tmp_p1 = []
            tmp_p2 = []
            for i in range(g + 1):
                tmp_p1.extend([generate_random_tropical_value(p_min, p_max)])
                tmp_p2.extend([generate_random_tropical_value(p_min, p_max)])

            self._p1 = tmp_p1
            self._p2 = tmp_p2

            result1 = tropical_matrix_0(self.n)
            result2 = tropical_matrix_0(self.n)
            for (el1, el2, i) in zip(self._p1, self._p2, range(self.g + 1)):
                result1 += (self.A ** i) * el1
                result2 += (self.B ** i) * el2

            self._p_1_A = result1
            self._p_2_B = result2

            self.m = None
            self._K = None

    def send_message(self, prnt=False):
        self.m = self._p_1_A * self._p_2_B
        if prnt:
            print("Message:\n" + str(self.m))
        return self.m

    def set_Key(self, v):
        self._K = self._p_1_A * v * self._p_2_B

    def get_Key(self):
        return self._K

    def check_Key(self, check_K):
        return check_K == self._K
